package org.nakedobjects.nof.util.memento;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.OneToManyAssociation;
import org.nakedobjects.noa.reflect.OneToOneAssociation;
import org.nakedobjects.noa.reflect.ValueAssociation;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.UnknownTypeException;


public class Memento implements Transferable, Serializable {
    private final static long serialVersionUID = 1L;
    private static final Logger LOG = Logger.getLogger(Memento.class);
    private Data state;

    /**
     * Creates a memento that hold the state for the specified object. This object is Serializable and can be
     * passed over the network easily. Also for a persistent only the references' OIDs are held, avoiding the
     * need for serializing the whole object graph.
     */
    public Memento(final NakedObject object) {
        state = object == null ? null : createData(object);
        LOG.debug("created memento for " + this);
    }

    private Data createData(final Naked object) {
        if (object instanceof NakedCollection) {
            return createCollectionData(object);
        } else {
            return createObjectData((NakedObject) object);
        }
    }

    private Data createCollectionData(final Naked object) {
        NakedCollection coll = (NakedCollection) object;
        Data[] collData = new Data[coll.size()];
        Enumeration elements = coll.elements();
        int i = 0;
        while (elements.hasMoreElements()) {
            NakedObject ref = (NakedObject) elements.nextElement();
            collData[i++] = createReferenceData(ref);
            // collData[i] = new Data(ref.getOid(), ref.getResolveState().name(), ref.getSpecification().getFullName());
        }
        return new CollectionData(coll.getOid(), NakedCollection.class.getName(), collData);
    }

    private ObjectData createObjectData(final NakedObject object) {
        NakedObjectSpecification cls = object.getSpecification();
        NakedObjectField[] fields = cls.getFields();
        ObjectData data = new ObjectData(object.getOid(), object.getResolveState().name(), cls.getFullName());
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].isPersisted()) {
                createFieldData(object, data, fields[i]);
            }
        }
        return data;
    }

    private void createFieldData(final NakedObject object, ObjectData data, NakedObjectField field) {
        Object fieldData;
        if (field.isCollection()) {
            NakedCollection coll = (NakedCollection) field.get(object);
            fieldData = createCollectionData(coll);
        } else if (field.isObject()) {
            NakedObject ref = (NakedObject) ((OneToOneAssociation) field).get(object);
            fieldData = createReferenceData(ref);
        } else if (field.isValue()) {
            NakedValue value = (NakedValue) field.get(object);
            fieldData = value.asEncodedString();
        } else {
            throw new UnknownTypeException(field);
        }
        data.addField(field.getId(), fieldData);
    }

    private Data createReferenceData(NakedObject ref) {
        Data fieldData;
        if (ref != null) {
            if (ref.getOid().isTransient()) {
                fieldData = createObjectData(ref);
            } else {
                String resolvedState = ref.getResolveState().name();
                String specification = ref.getSpecification().getFullName();
                fieldData = new Data(ref.getOid(), resolvedState, specification);
            }
        } else {
            fieldData = null;
        }
        return fieldData;
    }

    public Oid getOid() {
        return state.oid;
    }

    public NakedObject recreateObject() {
        if (state == null) {
            return null;
        } else {
            NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(state.className);
            NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
            NakedObject object;
            ResolveState targetState;
            if (getOid().isTransient()) {
                object = objectLoader.recreateTransientInstance(getOid(), spec);
                targetState = ResolveState.SERIALIZING_TRANSIENT;
            } else {
                object = objectLoader.recreateAdapterForPersistent(getOid(), spec);
                targetState = ResolveState.UPDATING;
            }
            updateObject(object, state, targetState);
            LOG.debug("recreated object " + object.getOid());
            return object;
        }
    }

    private NakedObject recreateReference(final Data data) {
        NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
        synchronized (objectLoader) {
            Oid oid = data.oid;
            NakedObject ref;
            if (oid == null) {
                ref = null;
            } else {
                NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(data.className);
                ResolveState resolveState;
                if (oid.isTransient()) {
                    ref = objectLoader.recreateTransientInstance(oid, spec);
                    resolveState = ResolveState.getResolveState(data.resolveState);
                } else {
                    ref = objectLoader.recreateAdapterForPersistent(oid, spec);
                    resolveState = ResolveState.GHOST;
                }
                if (data instanceof ObjectData) {
                    updateObject(ref, data, resolveState);
                }
            }
            return ref;
        }
    }

    public String toString() {
        return "[" + (state == null ? null : state.className + "/" + state.oid + state) + "]";
    }

    /**
     * Updates the specified object (assuming it is the correct object for this memento) with the state held
     * by this memento.
     * 
     * @throws IllegalArgumentException
     *             if the memento was created from different logical object to the one specified (i.e. its oid
     *             differs).
     */
    public void updateObject(final NakedObject object) {
        updateObject(object, state, ResolveState.RESOLVING);
    }

    private void updateObject(final NakedObject object, final Data state, final ResolveState resolveState) {
        Object oid = object.getOid();
        if (oid != null && !oid.equals(state.oid)) {
            throw new IllegalArgumentException("This memento can only be used to  update the naked object with the Oid "
                    + state.oid + " but is " + oid);

        } else {
            if (!(state instanceof ObjectData)) {
                throw new NakedObjectRuntimeException("Expected an ObjectData but got " + state.getClass());
            } else {
                NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
                updateObject(object, resolveState, state, objectLoader);
            }
            LOG.debug("object updated " + object.getOid());
        }
    }

    private void updateObject(
            final NakedObject object,
            final ResolveState resolveState,
            Data state,
            NakedObjectLoader objectLoader) {
        if (object.getResolveState().isValidToChangeTo(resolveState)) {
            objectLoader.start(object, resolveState);
            updateFields(object, state);
            objectLoader.end(object);
        } else if (object.getResolveState() == ResolveState.TRANSIENT && resolveState == ResolveState.TRANSIENT) {
            updateFields(object, state);
        } else {
            ObjectData od = (ObjectData) state;
            if (od.containsField()) {
                throw new NakedObjectRuntimeException("Resolve state (for " + object
                        + ") inconsistent with fact that data exists for fields");
            }
        }
    }

    private void updateFields(final NakedObject object, Data state) {
        ObjectData od = (ObjectData) state;
        NakedObjectField[] fields = object.getSpecification().getFields();
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].isPersisted()) {
                updateField(object, od, fields[i]);
            }
        }
    }

    private void updateField(final NakedObject object, ObjectData od, NakedObjectField field) {
        Object fieldData = od.getEntry(field.getId());
        if (field.isCollection()) {
            updateOneToManyAssociation(object, (OneToManyAssociation) field, (CollectionData) fieldData);
        } else if (field.isObject()) {
            updateOneToOneAssociation(object, (OneToOneAssociation) field, (Data) fieldData);
        } else if (field.isValue()) {
            NakedValue value = NakedObjectsContext.getObjectLoader().createValueInstance(field.getSpecification());
            value.restoreFromEncodedString((String) fieldData);
            ((ValueAssociation) field).initValue(object, value);
        }
    }

    private void updateOneToManyAssociation(
            final NakedObject object,
            final OneToManyAssociation field,
            final CollectionData collectionData) {
        NakedCollection collection = (NakedCollection) field.get(object);
        Vector original = new Vector();
        Enumeration elements = collection.elements();
        while (elements.hasMoreElements()) {
            original.addElement(elements.nextElement());
        }

        for (int j = 0; j < collectionData.elements.length; j++) {
            NakedObject element = recreateReference((Data) collectionData.elements[j]);
            if (!collection.contains(element)) {
                LOG.debug("  association " + field + " changed, added " + element.getOid());
                field.addElement(object, element);
            } else {
                field.removeElement(object, element);
            }
        }

        int size = original.size();
        for (int i = 0; i < size; i++) {
            NakedObject element = (NakedObject) original.elementAt(i);
            LOG.debug("  association " + field + " changed, removed " + element.getOid());
            field.removeElement(object, element);
        }
    }

    private void updateOneToOneAssociation(final NakedObject object, final OneToOneAssociation field, final Data fieldData) {
        if (fieldData == null) {
            field.initAssociation(object, null);
        } else {
            NakedObject ref = recreateReference(fieldData);
            if (field.get(object) != ref) {
                LOG.debug("  association " + field + " changed to " + ref.getOid());
                field.initAssociation(object, ref);
            }
        }
    }

    public void writeData(final TransferableWriter data) {
        data.writeObject(state);
    }

    public void restore(final TransferableReader data) {
        state = (Data) data.readObject();
    }

    public void debug(DebugString debug) {
        state.debug(debug);
    }
}
// Copyright (c) Naked Objects Group Ltd.
